from typing import Any, Callable
from pytest_benchmark.fixture import BenchmarkFixture  # type: ignore[import-untyped]
from gloomy import gloom
from glom import glom  # type: ignore[import-untyped]
from functools import partial
import pytest

from tests.utils import Obj


@pytest.mark.parametrize(
    ("impl"),
    [
        pytest.param(partial(glom, default=None), id="glom"),
        pytest.param(partial(gloom, default=None), id="gloom"),
        pytest.param(None, id="manual-impl"),
    ],
)
class TestBenchmark:
    def test_dict_key_missing(
        self,
        benchmark: BenchmarkFixture,
        impl: Callable | None,
    ):
        def _manual_impl(target: Any, spec: str):
            return target.get("missing")

        kwargs = dict(target={}, spec="missing")
        result = benchmark(impl or _manual_impl, **kwargs)
        assert result is None

    def test_dict_key_exists(
        self,
        benchmark: BenchmarkFixture,
        impl: Callable | None,
    ):
        def _manual_impl(target: Any, spec: str):
            return target.get("a", {}).get("b", {}).get("c")

        kwargs = dict(target={"a": {"b": {"c": 123}}}, spec="a.b.c")
        result = benchmark(impl or _manual_impl, **kwargs)
        assert result == 123

    def test_dict_key_exists_deep(
        self,
        benchmark: BenchmarkFixture,
        impl: Callable | None,
    ):
        def _manual_impl(target: Any, spec: str):
            try:
                return target["a"]["b"]["c"]["d"]["e"]["f"]["g"]["h"]["i"]
            except (TypeError, KeyError):
                return None

        data = {"a": {"b": {"c": {"d": {"e": {"f": {"g": {"h": {"i": 123}}}}}}}}}

        kwargs = dict(target=data, spec="a.b.c.d.e.f.g.h.i")
        result = benchmark(impl or _manual_impl, **kwargs)
        assert result == 123

    def test_obj_attr_missing(
        self,
        benchmark: BenchmarkFixture,
        impl: Callable | None,
    ):
        def _manual_impl(target: Any, spec: str):
            return getattr(target, "missing", None)

        kwargs = dict(target=Obj(), spec="missing")
        result = benchmark(impl or _manual_impl, **kwargs)
        assert result is None

    def test_obj_attr_exists_int(
        self,
        benchmark: BenchmarkFixture,
        impl: Callable | None,
    ):
        def _manual_impl(target: Any, spec: str):
            try:
                return target.a.b.c
            except AttributeError:
                return None

        kwargs = dict(target=Obj(a=Obj(b=Obj(c=123))), spec="a.b.c")
        result = benchmark(impl or _manual_impl, **kwargs)
        assert result == 123

    def test_obj_attr_list_element_missing(
        self,
        benchmark: BenchmarkFixture,
        impl: Callable | None,
    ):
        def _manual_impl(target: Any, spec: str):
            try:
                return target.a[0].b
            except (AttributeError, IndexError):
                return None

        kwargs = dict(target=Obj(a=[]), spec="a.0.b")
        result = benchmark(impl or _manual_impl, **kwargs)
        assert result is None

    def test_obj_attr_list_element_exists_int(
        self,
        benchmark: BenchmarkFixture,
        impl: Callable | None,
    ):
        def _manual_impl(target: Any, spec: str):
            try:
                return target.a[0].b
            except (AttributeError, IndexError):
                return None

        class ObjectB:
            b = 123

        class TargetObject:
            a = [ObjectB()]

        kwargs = dict(target=Obj(a=[Obj(b=123)]), spec="a.0.b")
        result = benchmark(impl or _manual_impl, **kwargs)
        assert result == 123



from typing import Any
from tests.utils import Obj
import pytest
from gloomy import gloom


@pytest.mark.parametrize(
    ("target", "spec", "expected"),
    [
        ([1], "0", 1),
        ({0: 1}, "0", 1),
        ({"0": 1}, "0", 1),
        ({"a": 123}, "a", 123),
        (Obj(a=123), "a", 123),
        ({"a": {"b": {"c": 123}}}, "a.b.c", 123),
        ({"a": Obj(b=Obj(c=123))}, "a.b.c", 123),
        ([{"li": [{"foo": "bar"}]}], "0.li.0.foo", "bar"),
    ],
)
def test_valid_paths(target: Any, spec: str, expected: Any):
    result = gloom(target, spec)
    assert result == expected



from typing import Any, Mapping, Sequence

_no_default = object()


class PathAccessError(Exception):
    """Raised when specified path (spec) cannot be accessed and no default is provided."""


def gloom(
    target: Sequence | Mapping | object | None,
    spec: str,
    default: Any = _no_default,
) -> Any:
    """
    Access a nested attribute, key or index of an object, mapping or sequence.

    Raises:
        PathAccessError: if the path cannot be accessed and no default is provided.

    """
    if target is None:
        if default is _no_default:
            msg = "Cannot access path as target is None."
            raise PathAccessError(msg)
        return default

    path_parts = spec.split(".")
    location = target

    for part in path_parts:
        # Get key/index of mapping/sequence
        if getitem := getattr(location, "__getitem__", None):
            if part.isdigit():
                try:
                    # Sequence or mapping with int keys
                    location = getitem(int(part))
                    continue
                except IndexError as e:
                    if default is _no_default:
                        raise PathAccessError from e
                    return default
                except KeyError:
                    # Possibly mapping with numeric string keys
                    pass
            try:
                location = getitem(part)
                continue
            except KeyError as e:
                if default is _no_default:
                    raise PathAccessError from e
                return default
        try:
            location = getattr(location, part)
        except AttributeError as e:
            if default is _no_default:
                raise PathAccessError from e
            return default

    return location
